/*
Portions of this code related to normal mapping 
and tangent space calculations derived from 
http://www.gamedev.net/community/forums/topic.asp?topic_id=504221

Blinn-Phong lighting technique from
http://en.wikipedia.org/wiki/Blinn%E2%80%93Phong_shading_model
*/

#include "globals.fxh"

float3 AmbientLightColor    = float3( 0.5f, 0.5f, 0.5f );

struct Lighting
{
    float3 Diffuse;
    float3 Specular;
};

struct PointLight
{
	float3 Position;
	float4 DiffuseColor;
	float  DiffusePower;
	float4 SpecularColor;
	float  SpecularPower;
};

cbuffer Lights
{
	PointLight Light4		: Light4;
}

float3 EyeFromViewIT(){
	return float3(ViewITMatrix[0].w,ViewITMatrix[1].w,ViewITMatrix[2].w);
}

Lighting GetPointLight(PointLight light, float3 viewDir, float3 normal, float3 worldPos, float3x3 WorldToTangentSpace)
{
	Lighting OUT;
	if(light.DiffusePower > 0)
	{
		// light vector in tangent space
		float3 lightVec  = light.Position - worldPos;
		lightVec = mul( WorldToTangentSpace, -lightVec );

		float dist = length(lightVec); // GET THE DISTANCE OF THIS VECTOR
		dist = dist * dist; // USES INVERSE SQUARE FOR DISTANCE ATTENUATION
		lightVec = normalize(lightVec); // NORMALIZE THE VECTOR

		// INTENSITY OF THE DIFFUSE LIGHT	
		// SATURATE TO KEEP WITHIN THE 0-1 RANGE
		// DOT PRODUCT OF THE LIGHT DIRECTION VECTOR AND THE SURFACE NORMAL
		float i = saturate(dot(lightVec, normal)); 

		OUT.Diffuse = i * light.DiffuseColor * light.DiffusePower / dist; // CALCULATE THE DIFFUSE LIGHT FACTORING IN LIGHT COLOUR, POWER AND THE ATTENUATION

		//CALCULATE THE HALF VECTOR BETWEEN THE LIGHT VECTOR AND THE VIEW VECTOR. THIS IS CHEAPER THAN CALCULATING THE ACTUAL REFLECTIVE VECTOR
		float3 h =  normalize(lightVec + viewDir);

		// INTENSITY OF THE SPECULAR LIGHT	
		// DOT PRODUCT OF NORMAL VECTOR AND THE HALF VECTOR TO THE POWER OF THE SPECULAR HARDNESS
		i = pow(saturate(dot(normal, h)), g_nMaterialShininess);
		OUT.Specular = i * light.SpecularColor * light.SpecularPower / dist; // CALCULATE THE SPECULAR LIGHT FACTORING IN LIGHT SPECULAR COLOUR, POWER AND THE ATTENUATION
	}
	return OUT;
}


////////////////////////////////////////////////
// Vertex Shader - Main Function
///////////////////////////////////////////////
PS_INPUT VS( VS_INPUT vsInput )
{
	PS_INPUT psInput;
	
	// transform vertex pos to homogeneous coordinates
    psInput.Pos = mul( vsInput.Pos, WVPMatrix );

	// pass texture coords
	psInput.Tex = vsInput.Tex;
	
	// world space to tangent space matrix
	float3x3 worldToTangentSpace = float3x3(	
    		mul( vsInput.Tangent,  WorldITMatrix ).xyz,
    		mul( vsInput.Binormal, WorldITMatrix ).xyz,
    		mul( vsInput.Normal,   WorldITMatrix ).xyz
	);

	psInput.WorldToTangentSpace = worldToTangentSpace;

	// vertex position in world space
	float3 posWorld = mul( vsInput.Pos, WorldMatrix );
	psInput.WorldPos = posWorld;

	psInput.SurfaceNormal = mul(vsInput.Normal, worldToTangentSpace);

    return psInput;
}


///////////////////////////////////////////////
// Pixel Shader
///////////////////////////////////////////////
float4 PS(PS_INPUT psInput) : SV_Target
{
	// view vector in tangent space
	float3 viewDir = EyeFromViewIT() - psInput.WorldPos;
	viewDir  = mul( psInput.WorldToTangentSpace, viewDir );

	float4 baseColor = DiffuseTexture.Sample( samLinear,  psInput.Tex );
	float3 normal = normalize( 2 * psInput.SurfaceNormal - 1 );
	
	viewDir = normalize( viewDir );
	
	// test hardcoded lights
	PointLight light1;
	light1.Position			= float3( -10.0f, 5.0f, 5.0f );
	light1.DiffuseColor		= float4( 1.0f, 0.0f, 0.0f, 1.0f );
	light1.SpecularColor	= float4( 1.0f, 1.0f, 1.0f, 1.0f );
	light1.SpecularPower	= 128.0;
	light1.DiffusePower		= 128;


	PointLight light2;
	light2.Position			= float3( 10.0f, 5.0f, 5.0f );
	light2.DiffuseColor		= float4( 0.0f, 0.0f, 1.0f, 1.0f );
	light2.SpecularColor	= float4( 1.0f, 1.0f, 1.0f, 1.0f );
	light2.SpecularPower	= 128.0;
	light2.DiffusePower		= 128;

	PointLight light3;
	light3.Position			= float3( 0.0f, 5.0f, -10.0f );
	light3.DiffuseColor		= float4( 0.0f, 1.0f, 0.0f, 1.0f );
	light3.SpecularColor	= float4( 1.0f, 1.0f, 1.0f, 1.0f );
	light3.SpecularPower	= 128.0;
	light3.DiffusePower		= 128;

	
	Lighting l1 = GetPointLight(light1, viewDir, normal, psInput.WorldPos, psInput.WorldToTangentSpace);
	Lighting l2 = GetPointLight(light2, viewDir, normal, psInput.WorldPos, psInput.WorldToTangentSpace);
	Lighting l3 = GetPointLight(light3, viewDir, normal, psInput.WorldPos, psInput.WorldToTangentSpace);
	Lighting l4 = GetPointLight(Light4, viewDir, normal, psInput.WorldPos, psInput.WorldToTangentSpace);

	Lighting final;
	final.Diffuse = l1.Diffuse + l2.Diffuse + l3.Diffuse + l4.Diffuse;
	final.Specular = l1.Specular + l2.Specular + l3.Specular + l4.Specular;
	
	return float4( saturate( 
	(AmbientLightColor * baseColor.rgb) +
	(final.Diffuse * baseColor.rgb * g_vMaterialDiffuse * 0.7) + 
	(final.Specular * g_vMaterialSpecular * 0.9) )
	, ( baseColor.a * g_fMaterialAlpha ) );
	
}


// Define the technique
technique10 Render
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS() ) );
    }
}
